#---------------------Cartogram Utilities (cartUtils)
# Purpose:     a set of utilities for cartogram construction
#
# Author:      Barry J Kronenfeld
#
# Created:     2014
# Copyright:   (c) bjkronenfeld 2014
# Licence:     GNU Public License v. 3.0
#-------------------------------------------------------------------------------
import arcpy
import arcpy.da
arcpy.env.overwriteOutput=True
# toggle run mode
run_from_arcmap=True

def report(txtMessage):
    print txtMessage
    arcpy.AddMessage(txtMessage)

def validateInput(origSpace, cartSpace, inData, outWS):
    # makes sure the input parameters are valid
    if not arcpy.Exists(origSpace):
        arcpy.AddMessage("File does not exist: " + origSpace)
        exit
    if not arcpy.Exists(cartSpace):
        arcpy.AddMessage("File does not exist: " + cartSpace)
        exit
    for ds in inData:
        if not arcpy.Exists(ds):
            arcpy.AddMessage("File does not exist: " + ds)
        exit
    if not arcpy.Exists(outWS):
        arcpy.AddMessage("Workspace does not exist: " + outWS)
        exit

def clearMemory():
    arcpy.Delete_management("in_memory")

def vertexcoordinatelist(geo):
    # creates a list of coordinate pairs (as lists) corresponding to the given geometry
    # set up arrays
    v=[]
    # get vertex array
    vertexarray=geo.getPart(0)
    # loop through vertices
    for j in range(vertexarray.count):
        vertex=vertexarray.getObject(j)
        # add x-y coordinate pair to result list
        v.append([vertex.X,vertex.Y])
    # return list
    return v
def barycentricCoordinates(p,v):
    # returns three barycentric coordinates
    # input point p should be list of two numbers
    # input triangle vertices v should be list of triangle vertex coordinate pairs
    # example:
    # p=[1.0,1.0]
    # v=[[0.0,0.0],[0.0,2.0],[2.0,0.0],[0.0,0.0]

    # get total area of triangle
    totalArea=polyArea(v)

    # get vertices
    v0=v[0]
    v1=v[1]
    v2=v[2]
    # create triangle 0 and calculate its area
    t0=[p,v1,v2,p]
    a0=polyArea(t0)
    # create triangle 1 and get its area
    t1=[p,v2,v0,p]
    a1=polyArea(t1)
    # calculate coordinates
    c0=a0/totalArea
    c1=a1/totalArea
    c2=1-c0-c1
    # return results
    return [c0,c1,c2]
def euclideanCoordinates(b,v):
    # converts from barycentric coordinates to Euclidean coordinates
    # b should be list of three barycentric coordinates summing to 1
    # v should be list of three coordinate pairs
    # example:
    # b=[0.5,0.3,0.2]
    # v[[0.0,0.0],[0.0,2.0],[2.0,0.0],[0.0,0.0]]
    v0=v[0]
    v1=v[1]
    v2=v[2]
    x0=v0[0]
    x1=v1[0]
    x2=v2[0]
    y0=v0[1]
    y1=v1[1]
    y2=v2[1]
    rX=b[0]*x0+b[1]*x1+b[2]*x2
    rY=b[0]*y0+b[1]*y1+b[2]*y2
    return [rX,rY]
def polyArea(vertexList):
    # returns the area of a polygon
    # input vertices should be list of coordinate pairs
    #  with first vertex repeated
    # example:
    # vertexList=[[0.0,0.0],[0.0,2.0],[2.0,0.0],[0.0,0.0]]
    a=0
    for i in range(len(vertexList)-1):
        v0=vertexList[i]
        v1=vertexList[i+1]
        x0=v0[0]
        y0=v0[1]
        x1=v1[0]
        y1=v1[1]
        a+=(x1-x0)*(y0+y1)/2
    return a


class c_transformer():
    # a class to define a transformation from
    # one set of triangles to another
    base=None
    def __init__(self, baseFC, targetFC):
        self.baseFC=arcpy.CopyFeatures_management(baseFC,r'in_memory\base')
        self.targetFC=arcpy.CopyFeatures_management(targetFC,r'in_memory\target')
        self.snarf_up_input_data()


    def snarf_up_input_data(self):
        # validates data, appends ID field, and extracts coordinates
        # specifically, checks the input triangle feature classes to make sure that:
        # 1. they are polygon feature classes
        # 2. they have the same number of features
        # 3. all features have only one part
        # 4. all features have exactly 4 vertices
        # NOTE: This doesn't check topology (yet...)

        report("processing transformation transformation data...")
        # set up lists for feature x and y coordinates
        self.basecoord=[]
        self.targetcoord=[]
        # process both source and target datasets
        featCount=[]
        for i in range(2):
            # get reference to dataset
            ds=""
            if i==0:
                ds=self.baseFC
            else:
                ds=self.targetFC
            # make sure it's a shapefile or feature class
            desc=arcpy.Describe(ds)
            dt = desc.dataType
            if dt<>"ShapeFile" and dt<>"FeatureClass" and dt<>"FeatureLayer":
                arcpy.AddError("File {} ({}) is not a shapefile or feature class.".format(ds,dt))
                raise TypeError
            # make sure it's a polygon feature class
            shpType=desc.shapeType
            if shpType<>"Polygon":
                msg="Feature class " + ds + " does not contain polygons."
                print msg
                arcpy.AddError(msg)
                raise TypeError
            # append triangle ID field
            report("indexing triangles...")
            fields=desc.fields
            has_triangle_field=False
            for f in fields:
                if f.name=="triangle":
                    if f.type == "Integer":
                        has_triangle_field=True
                    else:
                        arcpy.DeleteField_management(ds,"triangle")
            if not has_triangle_field:
                arcpy.AddField_management(ds,"triangle", "Integer")
            tri_id=0
            ud = arcpy.da.UpdateCursor(ds,["triangle"])
            for row in ud:
                row[0]=tri_id
                ud.updateRow(row)
                tri_id+=1
            del ud
            # get coordinates of each triangle as list of lists and put in a giant list
            sc=arcpy.da.SearchCursor(ds,["SHAPE@XY"], explode_to_points=True)
            vcount=0
            coords=[]
            featnum=0
            numfeat=arcpy.GetCount_management(ds).getOutput(0)
            report("extracting triangle coordinates ({})...".format(i))
            for row in sc:
                if featnum % 100000 == 0 and vcount==0 and featnum > 0:
                    report("... {} of {}".format(str(featnum),str(numfeat)))
                vcount=vcount+1
                xy=row[0]
                coords.append([xy[0],xy[1]])
                if vcount==4:
                    vcount=0
                    featnum+=1
                    if i==0:
                        self.basecoord.append(coords)
                    else:
                        self.targetcoord.append(coords)
                    coords=[]
        # check that number of triangles match
        if len(self.basecoord)<>len(self.targetcoord):
            print "Number of features don't match!"
            arcpy.AddError(self.baseFC + " and " + self.targetFC + " have different numbers of features.")
            raise ValueError
        # OK, here's where we would start checking topology...
        print "Mesh coordinates extracted successfully."

    def transformCoordinates(self, fc):
        # Applies coordinate transformation to every vertex in the input feature class (fc)
        # Warning: This modifies the input feature class (fc)
        try:
            # get ID field
            desc=arcpy.Describe(fc)
            idfield=desc.OIDFieldName
            # get all points of input feature class as geometries
            report("...extracting vertex coordinates")
            sc=arcpy.da.SearchCursor(fc,["SHAPE",idfield], explode_to_points=True)
            geoList=[]
            idList=[]
            for row in sc:
                geoList.append(row[0])
                idList.append(row[1])
            del sc
            n=len(geoList)
            report("..." + str(len(geoList)) + " vertices found")
            # determine triangle of each point
            report("...identifying containing triangles")
            vertex_triangle=[]
            vertexfc=arcpy.FeatureVerticesToPoints_management(fc,r'in_memory\featvert')
            fields=desc.fields
            for f in fields:
                if f.name=="triangle":
                    arcpy.DeleteField_management(vertexfc,"triangle")
            vertexjoin=arcpy.SpatialJoin_analysis(vertexfc,self.baseFC,r'in_memory\vertexjoin',match_option="WITHIN")
            for row in arcpy.da.SearchCursor(vertexjoin,["triangle"]):
                if row[0]==None:
                    vertex_triangle.append(-1)
                else:
                    vertex_triangle.append(row[0])
            arcpy.Delete_management(r'in_memory\featvert')
            arcpy.Delete_management(r'in_memory\vertexjoin')
            # create update cursor with original point vertices
            uc=arcpy.da.UpdateCursor(fc,"SHAPE",explode_to_points=True)
            # convert coordinates
            report("...transforming to cartogram space")
            # loop through rows in cursor
            ptID=0
            initpoly=[]
            lastFID=-1
            for row in uc:
                # get original triangle
                curpoly=vertex_triangle[ptID]
                # only need to alter coordinates if it was in a triangle:
                if curpoly > -1:
                    # get original point
                    geo=geoList[ptID]
                    # put point coordinates into a list
                    p=[geo[0],geo[1]]
                    # get triangle coordinates
                    basev=self.basecoord[curpoly]
                    targetv=self.targetcoord[curpoly]
                    # get barycentric coordinates
                    b=barycentricCoordinates(p,basev)
                    # get transformed coordinates
                    xy=euclideanCoordinates(b,targetv)
                    point = arcpy.Point(xy[0],xy[1])
                    # update feature with new coordinates
                    ptgeo=arcpy.PointGeometry(point)
                    row[0]=ptgeo
                    uc.updateRow(row)
                ptID+=1
                if ptID%10000==0:
                    report("...finished " + str(ptID) + " of " + str(n) + " points")
            del uc
        except Exception as e:
            print e.message

def numVertices(fc):
    # counts the number of vertices in a feature class
    sc=arcpy.da.SearchCursor(fc,["SHAPE"], explode_to_points=True)
    r=0
    for row in sc:
        r+=1
    del sc
    return r

def transformData(srcLine, origSpace, cartSpace, inData, outWS, guarantee_topology):
    # transforms data from a space defined by one set of triangles
    # to the space defined by another
    # for now, all input & output must be shapefiles

    # validate input
    validateInput(origSpace,cartSpace,inData,outWS)
    # make copy of original space data in output workspace,
    # for use with Integrate function (to add vertices where features
    # intersect triangle edges)
    report('reading mesh features...')
    arcpy.env.workspace=outWS
    arcpy.env.overwriteOutput=True
    origTriName=r'in_memory\originTriangles'
    originTriangles= arcpy.management.CopyFeatures(origSpace,origTriName)
    #    report('extracting edges...')
    #    origEdgeName=r'in_memory\orig_edge'
    #    orig_edge=arcpy.FeatureToLine_management(originTriangles,origEdgeName)

    # set working coordinate system to mesh triangle coordinate system
    orig_desc = arcpy.Describe(origTriName)
    sr=orig_desc.spatialReference
    arcpy.env.outputCoordinateSystem=sr
    # create transformation object
    transformer=c_transformer(origSpace,cartSpace)
    # loop through datasets
    for origds in inData:
        # get name of layer
        desc=arcpy.Describe(origds)
        dsname=desc.baseName
        if dsname[-4:]==".shp":
            dsname=dsname[:-4] + "_cartogram.shp"
        else:
            dsname=dsname + "_cartogram"
        report("Working on " + dsname + "...")
        # make copy of the input feature class in the output workspace
        report("...copying features")
        newds=arcpy.management.CopyFeatures(origds,dsname).getOutput(0)
##        report(str(type(newds)))
##        report(newds)
        # for line and polygon feature classes, add vertices at intersections with origin triangles
        if guarantee_topology:
            desc = arcpy.Describe(newds)
            st = desc.shapeType
            if st=="Polyline" or st == "Polygon" or st == "MultiPatch":
                # integrate tool only seems to work with feature class
                # names in the environment workspace
                lineCopy = arcpy.CopyFeatures_management(srcLine,r'in_memory\srcLineCopy')
                ws_desc=arcpy.Describe(arcpy.env.workspace)
                ws_type=ws_desc.workspaceType
                ext=""
                if ws_type=="FileSystem":
                    ext=".shp"
                infeat=newds + " #;" + r'in_memory\srcLineCopy' + " #"
##                report(infeat)
##                report("...adding vertices on triangle edges")

                count1=numVertices(newds)
##                count2=numVertices(newds)
##                report(arcpy.env.workspace)
##                report(dsname)
##                report(newds)
##                report("Before: {},{}".format(str(count1),str(count2)))

                #arcpy.management.Integrate(infeat)


                arcpy.Integrate_management(r"'in_memory\srcLineCopy' 1;'" + newds + "' 2", '#')

##                count1=numVertices(dsname)
                count2=numVertices(newds)
                diff=count2-count1
                report("...inserted {} vertices at intersections with transformation mesh".format(str(diff)))


        # run the transformation
        transformer.transformCoordinates(newds)
        # add to map
        if run_from_arcmap:
            mxd=arcpy.mapping.MapDocument("CURRENT")
            df=arcpy.mapping.ListDataFrames(mxd,"*")[0]
            resultlyr=arcpy.mapping.Layer(newds)
            arcpy.mapping.AddLayer(df,resultlyr,"TOP")
            arcpy.RefreshActiveView()
            arcpy.RefreshTOC()
            arcpy.RefreshCatalog(newds)
        # report finish
        report("Finished " + dsname + ".")
    # delete unnecessary feature classes
    arcpy.management.Delete(originTriangles)

if __name__ == '__main__':
    report("starting...")
    # get input parameters
    transform_ws=r'C:\Box Sync\programming\ArcGIS Scripts\CartogramTransform\new2\transformation'

    ds1=r'C:\Box Sync\programming\ArcGIS Scripts\CartogramTransform\new2\USA_48states.shp'

    inData=[ds1]

    outWorkSpace=r'C:\Box Sync\programming\ArcGIS Scripts\CartogramTransform\TestData\USA\test_results.gdb'

    guarantee_topology = True

    if run_from_arcmap:
        transform_ws = arcpy.GetParameterAsText(0)
        inData = arcpy.GetParameter(1)
        outWorkSpace=arcpy.GetParameterAsText(2)
        guarantee_topology = arcpy.GetParameter(3)

    # get transformation files
    ws_desc=arcpy.Describe(outWorkSpace)
    ws_type=ws_desc.workspaceType
    ext=""
    if ws_type=="FileSystem":
        ext=".shp"
    arcpy.env.workspace=transform_ws
    ws_fc = arcpy.ListFeatureClasses()
    srcPoly=""
    srcLine=""
    trgPoly=""

    for fc in ws_fc:
        if 'source_poly' in fc:
            srcPoly=transform_ws + '\\' + fc
        if 'source_line' in fc:
            srcLine=transform_ws + '\\' + fc
        if 'target_poly' in fc:
            trgPoly=transform_ws + '\\' + fc

    if srcPoly=="" or trgPoly == "":
        arcpy.AddError("Error: cannot find transformation mesh files. The transformation workspace should contain two polygon feature classes labelled 'source_poly' and 'target_poly'.")

    transformExists=True
    if not arcpy.Exists(srcPoly):
        transformExists=False
    if not arcpy.Exists(srcLine):
        transformExists=False
    if not arcpy.Exists(trgPoly):
        transformExists=False

    if transformExists:
        # run transformation
        transformData(srcLine, srcPoly,trgPoly,inData,outWorkSpace, guarantee_topology)
        # report finish
        report("finished!")
    else:
        report("srcLine: {}".format(srcLine))
        report("srcPoly: {}".format(srcPoly))
        report("trgPoly: {}".format(trgPoly))
        arcpy.AddError("Transformation files not found. Transformation workspace (folder or geodatabase) must contain three feature classes named source_line, source_poly and target_poly.")